#!/usr/bin/perl

#
# vdradmin.pl by Thomas Koch <tom@linvdr.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# Or, point your browser to http://www.gnu.org/copyleft/gpl.html
#
# 08.10.2001
#

my($BASENAME);
BEGIN {
	$0 =~ /(^.*\/)/;
	$BASENAME = $1;
	unshift(@INC, $BASENAME . "lib/");
	unshift(@INC, $BASENAME);
}

use CGI qw/:no_debug/;
use IO::Socket;
use HTML::Template();
use Time::Local qw(timelocal);
use POSIX qw(strftime);
use MIME::Base64();

use strict;

my %CONFIG;
$CONFIG{VDR_HOST}         = "video";
$CONFIG{VDR_PORT}         = 2001;
$CONFIG{USERNAME}         = "linvdr";
$CONFIG{PASSWORD}         = "linvdr";
$CONFIG{LANGUAGE}         = "Deutsch";
$CONFIG{LOGLEVEL}         = 32783;
$CONFIG{CACHE_TIMEOUT}    = 5;
$CONFIG{CACHE_LASTUPDATE} = 0;
$CONFIG{AT_FUNC}          = 0;
$CONFIG{AT_TIMEOUT}       = 10; 
$CONFIG{AT_LIFETIME}      = 99;
$CONFIG{AT_PRIORITY}      = 99;
$CONFIG{TM_LIFETIME}      = 99;
$CONFIG{TM_PRIORITY}      = 99;
$CONFIG{TM_MARGIN_BEGIN}  = 10;
$CONFIG{TM_MARGIN_END}    = 10;
#
$CONFIG{LOGFILE}          = "vdradmind.log";
$CONFIG{SERVERPORT}       = 8001;
$CONFIG{RECORDINGS}       = 1;


my $VERSION               = "0.19";
my $SERVERVERSION         = "vdradmind/$VERSION";
my $TEMPLATEDIR           = "${BASENAME}template";
my $CONFFILE              = "${BASENAME}vdradmind.conf";
my $LOGFILE               = "${BASENAME}$CONFIG{LOGFILE}";
my $PIDFILE               = "${BASENAME}vdradmind.pid";
my $AT_FILENAME           = "${BASENAME}vdradmind.at";
my $I18NFILE							= "i18n.pl";

my($DEBUG) = 0;
my(%EPG, @CHAN, $SOCKET, $q); 
my(%mimehash) = ( 
  html => "text/html",
  png  => "image/png",
  gif  => "image/gif",
  jpg  => "image/jpeg"
);

sub true          () { 1 };
sub false         () { 0 };
sub CRLF          () { "\r\n" };
sub LOG_ACCESS    () { 1 };
sub LOG_SERVERCOM () { 2 };
sub LOG_VDRCOM    () { 4 };
sub LOG_STATS     () { 8 };
sub LOG_AT        () { 16 };
sub LOG_DEBUG     () { 32768 };

$SIG{INT} = sub { unlink($PIDFILE); exit(0) };

my($DAEMON) = 1;
for(@ARGV) {
  if(/-h|--help/) {
    print("Usage $0 [OPTION]...\n");
    print("A perl client for the Linux Video Disk Recorder.\n\n");
    print("  -nf  --nofork   don't fork\n");
    print("  -c   --config   run configuration dialog\n");
    print("  -k   --kill     kill a fork'ed vdradmin\n");
    print("  -h   --help     this here\n");
    print("\nReport bugs to <tom\@linvdr.org>.\n");
    exit(0);
  }
  if(/--nofork|-nf/) { $DAEMON = 0; last; }
  if(/--config|-c/) {
    $CONFIG{VDR_HOST} = 
      Question("What's your VDR hostname (e.g video.intra.net)?", "video");
    $CONFIG{VDR_PORT} = 
      Question("What's the port VDR listen to SVDRP query's?", "2001");
    $CONFIG{SERVERPORT} = 
      Question("On which port should vdradmin answer?", "8001");
    $CONFIG{USERNAME} = Question("Username?", "linvdr");
    $CONFIG{PASSWORD} = Question("Password?", "linvdr");
    
    open(CONF, ">$CONFFILE") || die "Cannot open $CONFFILE: $!\n";
    for(keys(%CONFIG)) {
      print(CONF "$_ = $CONFIG{$_}\n");
    }
    close(CONF);

    print("Config file sucessfull written.\n");
  }
  if(/--kill|-k/) {
    kill(2, getPID($PIDFILE));
    exit(0);
  }
}

if(-e "$PIDFILE") {
	print "There's already an copy of this program running! (pid: " . getPID($PIDFILE) . ")\n";
	print "If you feel this is a error, remove $PIDFILE!\n";
	exit(0);
}

if($DAEMON) {
  my($pid) = fork;
  if($pid != 0) {
    print("vdradmind.pl $VERSION started with pid $pid.\n");
    open(F, ">${BASENAME}vdradmind.pid"); print(F $pid); close(F);
    exit(0);
  }
}


ReadConfig();
my($Socket) = IO::Socket::INET->new( 
  Proto => 'tcp',
  LocalPort => $CONFIG{SERVERPORT},
  Listen => SOMAXCONN,
  Reuse => 1
);				     
die("can't start server: $!\n") if (!$Socket);
$Socket->timeout($CONFIG{AT_TIMEOUT} * 60) if($CONFIG{AT_FUNC});
$CONFIG{CACHE_LASTUPDATE} = 0;

##
# Mainloop
##
my($Client, $MyURL, $Referer, $Request, $Query);
while(1) {
  $Client = $Socket->accept();
	
	#
	if(!$Client) {
		UptoDate(1);
		next;
	}
	
	my $peer = $Client->peerhost;	
	my @Request = ParseRequest($Client);
	my $raw_request = $Request[0];	

	if($raw_request =~ /^GET (\/[\w\.\/-\:]*)([\?[\w=&\.\+\%-\:]*]*) HTTP\/1.\d$/) {
		($Request, $Query) = ($1, $2 ? substr($2, 1, length($2)) : undef);
	} else {
		print "no match HELP!\n";
		die;
	}
	
	my($Host, $username, $password, $http_useragent);
	for my $line (@Request) {
		if($line =~ /Referer: (.*)/) {
			$Referer = $1;
		}
		if($line =~ /Host: (.*)/) {
			$Host = $1;
		}
		if($line =~ /Authorization: basic (.*)/i) {
			($username, $password) = split(":", MIME::Base64::decode_base64($1), 2);
		}
		if($line =~ /User-Agent: (.*)/i) {
			$http_useragent = $1;
		}
	}
				
	# authenticate
	if($CONFIG{USERNAME} ne $username || $CONFIG{PASSWORD} ne $password) {
		headerNoAuth();
		close($Client);
		next;
	}
	
	# serve request
	my ($http_status, $bytes_transfered);
	$MyURL = "." . $Request;
	if($Request eq "/vdradmin.pl") { 
		$q = CGI->new($Query);
		my $aktion;
		for(qw(prog_detail prog_list timer_list at_timer_edit at_timer_new 
			at_timer_save at_timer_delete timer_new_form timer_add timer_delete 
			prog_summary rec_list rec_delete rec_detail show_top conf_list 
			prog_switch rc_show rc_hitk)) {
			($aktion = $q->param("aktion")) if($q->param("aktion") eq $_);
		}
		if($aktion) {
			eval("(\$http_status, \$bytes_transfered) = $aktion();");
		} else {
			#XXX
			exit(256);
		}
	} elsif($Request eq "/") {
		my $template = TemplateNew("index.html");
		($http_status, $bytes_transfered) = header("200", "text/html", $template->output); 
	} else {
		($http_status, $bytes_transfered) = SendFile($Request);
	}
	
	Log(LOG_ACCESS, access_log($Client->peerhost, $username, time(), $raw_request, 
		$http_status, $bytes_transfered, undef, $http_useragent));

	close($Client)
}

#############################################################################
#############################################################################
sub GetChannelDesc {
  my(%hash);
  for(@CHAN) {
    $hash{$_->{id}} = $_->{name};
  }
  return(%hash);
}

sub GetChannelDescByNumber {
  my $vdr_id = shift;

  if($vdr_id) {
    for(@CHAN) {
      if($_->{vdr_id} == $vdr_id) {
				return($_->{name});
      }
    }
  } else { return(0); }
}

sub ReadFile {
  my $file = shift;
  return if(!$file);
  
  my $buf;
  open(I18N, $file) || HTMLError("Cannot open $file!");
  while(<I18N>) {
    $buf .= $_;
  }
  return($buf);
}

sub FullDay {
  my @I18N_Days;
  eval(ReadFile("$TEMPLATEDIR/$CONFIG{LANGUAGE}/$I18NFILE"));
  return($I18N_Days[shift]);
}
 
sub GetChannelID {
  my($sid) = $_[0];
  for(@CHAN) { 
    if($_->{id} == $sid) {
      return($_->{number});
    }
  }
}

sub EURL {
  my($text) = @_;
  $text =~ s/([^0-9a-zA-Z])/sprintf("%%%2.2x", ord($1))/ge;
  return($text);
}

sub HTMLError {
  my($error) = @_;   
  my($template) = HTML::Template->new(
    filename => "$TEMPLATEDIR/$CONFIG{LANGUAGE}/error.html");
  $template->param(error => $error);
  return(header("200", "text/html", $template->output));
  return;
}


sub FillInZero {
  my($str, $length) = @_;
  while(length($str) < $length) {
    $str = "0$str";
  }
  return($str);
}

sub ChanTree {
  undef(@CHAN);
  for(SendCMD(1, "lstc")) {
    chomp;
    my($vdr_id, $temp) = split(/ /, $_, 2);	
    my($name, $frequency, $polarization, $source, $symbolrate, $vpid, $apid,
      $tpid, $ca, $service_id) = split(/\:/, $temp);
    push(@CHAN, {
      vdr_id       => $vdr_id,
      name         => $name,
      frequency    => $frequency,
      polarization => $polarization,
      source       => $source,
      symbolrate   => $symbolrate,
      vpid         => $vpid,
      apid         => $apid,
      tpid         => $tpid,
      ca           => $ca,
      service_id   => $service_id
    }); 
  }
}

sub get_vdrid_from_channelid {
	my $channel_id = shift;
	if($channel_id =~ /^(\d*)$/) { # vdr 1.0.x & >= vdr 1.1.15
		for my $channel (@CHAN) {
			if($channel->{service_id} == $1) {
				return($channel->{vdr_id});
			}
		}
	} elsif($channel_id =~ /^(.*)-.*-(.*)-(.*)$/) {
		for my $channel (@CHAN) {
			if($channel->{source} eq $1 && $channel->{frequency} == $2 && $channel->{service_id} == $3) {
				return($channel->{vdr_id});
			}
		}	
	} else {
		print "Help!\n";
	}
}

sub get_name_from_vdrid {
	my $vdr_id = shift;
	if($vdr_id) {
		for my $channel (@CHAN) {
			if($channel->{vdr_id} == $vdr_id) {
				return($channel->{name});
			}
		}	
	}
}


#############################################################################
# EPG functions
#############################################################################

sub EPG_getEntry {
	my $vdr_id = shift;
	my $epg_id = shift;
	if($vdr_id && $epg_id) {
		for(@{$EPG{$vdr_id}}) {
			if($_->{id} == $epg_id) {
				return($_);
			}
		}
	}
}

sub EPG_buildTree {
  my(@DATA);
  for(SendCMD(1, "lste")) { push(@DATA, $_); }
  my($i, @events); undef(%EPG); my $id  = 0; my $bc = 0; 
  for($i = 0; $i <= $#DATA; $i++) {
    $_ = $DATA[$i];
    chomp;
    if(/^C ([^ ]+) *(.*)/) {
      $bc++;
      undef(@events);
      my($channel_id, $channel_name) = ($1, $2);
			my $vdr_id = get_vdrid_from_channelid($channel_id);
			#print "$channel_id, $channel_name, $vdr_id\n";
      for(; $i <= $#DATA; $i++) {
				$_ = $DATA[$i];
				if(/^E (.*) (.*) (.*) (.*)/ || /^E (.*) (.*) (.*)/) {
					my($time, $duration) = ($2, $3);
					my($title, $subtitle, $summary);
					for(; $i <= $#DATA; $i++) {
						$_ = $DATA[$i];
						if(/^T (.*)/) { $title = $1 }
						if(/^S (.*)/) { $subtitle = $1 }
						if(/^D (.*)/) { $summary = $1 }
						if(/^e/) {
							#
							push(@events, {
								channel_name => $channel_name,
								start        => $time,
								stop         => $time + $duration,
								duration     => $duration,
								title        => $title,
								subtitle     => $subtitle,
								summary      => $summary,
								id           => $id,
								vdr_id       => $vdr_id
							});
							$id++;
							last;
						}
					}
				} elsif(/^c/) {
					my($last) = 0;
					my(@temp);
					for(sort({ $a->{start} <=> $b->{start} } @events)) {
						next if($last == $_->{start});
						push(@temp, $_);
						$last = $_->{start};
					}
					$EPG{$vdr_id} = [ @temp ];
					last;
				}
			}
    }
  }
  Log(LOG_STATS, "EPGTree: $id events, $bc broadcasters");
}


#############################################################################
# Socket functions
#############################################################################

sub PrintToClient {
  my $string = join("", @_);
  return if(!defined($string));
  print($Client $string) if($Client);
}

sub ParseRequest {
	my $Socket = shift;
	my (@Request, $Line);
	do {
		$Line = <$Socket>;
		$Line =~ s/\r\n//g;
		push(@Request, $Line);
	} while($Line);
	return(@Request);
}

sub CloseSocket {
  SendLine($SOCKET, "quit\r\n");
  my $tmp;
  $tmp = <$SOCKET>;
  close($SOCKET); 
  undef($SOCKET);
}

sub SendLine {
  my($Socket, $cmd) = @_;
  print("SendLine: $cmd") if($DEBUG);
  my($result) = send($Socket, $cmd, 0);
  if($result != length($cmd)) {
    HTMLError("Error sending command to host $CONFIG{VDRHOST}\n");
  }
}

sub SendCMD {
  my($pc) = shift;
  my($cmd) = @_; 
  my($error) = 0;
  
  my($fc) = 0;
  if($SOCKET == undef) {
    $SOCKET = IO::Socket::INET->new(
      PeerAddr => $CONFIG{VDRHOST}, 
      PeerPort => $CONFIG{VDRPORT}, 
      Proto => 'tcp') || ($error = 1);
      $fc = 1;
  }
  if($error) {
      HTMLError(
				"can't connect to host \"$CONFIG{VDRHOST}\" port $CONFIG{VDRPORT}!");
      return(1);
  }
    
  # if its the first connection, read greeting...
  my($tmp);
  if($fc) {
    $tmp = <$SOCKET>;
  }

  print("SendCMD: $cmd") if($DEBUG);
  
  SendLine($SOCKET, "$cmd\r\n");
  my(@output); 
  my($first) = 1; my($last);
  while(<$SOCKET>) {
    if($first) { $last = substr($_, 0, 3); }
    s/\r//g;
    if(substr($_, 3, 1) eq "-") {
      s/^$last-//;
      push(@output, $_);
    } else {
      s/^$last //;
      push(@output, $_);
      last;
    }
  }
  
  if($DEBUG) {
    for(@output) {
      print "A: $_";
    }
  }
  
  #
  if(!$pc) {
    print("SendCMD: Closing connection.\n") if($DEBUG);
    CloseSocket();
  }
  
  return(@output);
}

sub headerTime {
  my(@weekdays) = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun");
  my(@months) = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");	     
  my($sec,$min,$hour,$day,$month,$year,$wday,$yday,$isdst) = gmtime(time);
  return(sprintf "%s, %s %s %s  %02d:%02d:%02d GMT", $weekdays[$wday], $day, $months[$month], $year, $hour, $min, $sec);
}

sub header {
	my($status, $ContentType, $data) = @_;
  PrintToClient("HTTP/1.0 $status\r\n");
  PrintToClient("Date: ", headerTime(), "\r\n");
  PrintToClient("Expires: Mon, 26 Jul 1997 05:00:00 GMT\r\n");
  PrintToClient("Server: $SERVERVERSION\r\n");
  PrintToClient("Content-Length: ", length($data), "\r\n") if($data);
  PrintToClient("Connection: close\r\n");
  PrintToClient("Content-type: $ContentType\r\n\r\n") if($ContentType);
  PrintToClient($data) if($data);
	return($status, length($data));
}

sub headerForward {
  my $url = shift;
  PrintToClient("HTTP/1.0 302 Found\r\n");
  PrintToClient("Date: ", headerTime(), "\r\n");
  PrintToClient("Server: $SERVERVERSION\r\n");
  PrintToClient("Connection: close\r\n");
  PrintToClient("Location: $url\r\n");
  PrintToClient("Content-type: text/html\r\n\r\n");
	return(302, 0);
}

sub headerNoAuth {
  my $template = TemplateNew("noauth.html");
	my $data = $template->output;
  PrintToClient("HTTP/1.0 401\r\n");
  PrintToClient("Date: ", headerTime(), "\r\n");
  PrintToClient("Server: $SERVERVERSION\r\n");
  PrintToClient("WWW-Authenticate: Basic realm=\"vdradmind\"\r\n");
  PrintToClient("Content-Length: ", length($data), "\r\n") if($data);
  PrintToClient("Connection: close\r\n");
  PrintToClient("Content-type: text/html\r\n\r\n");
  PrintToClient($data);
  return(401, length($data));
}


sub SendFile {
  sub FileError {
    my $template = HTML::Template->new(
      filename => "$TEMPLATEDIR/$CONFIG{LANGUAGE}/noperm.html");
    $template->param(
      title => $_[0],
      h1    => $_[1],
      error => $_[2],
    );
    return(header("$_[0] $_[1]", "text/html", $template->output));
  }
  my($File) = @_;
  my($buf, $temp);
  $File =~ s/^\///;
	my $FileWithPath = $BASENAME . $File;
  if(-e $FileWithPath) {
    if(-r $FileWithPath) {
      $buf = ReadFile($FileWithPath);
			$temp = $File;
      $temp =~ /([A-Za-z0-9]+)\.([A-Za-z0-9]+)/;
      if(!$mimehash{$2}) { die("can't find mime-type \'$2\'\n"); }
      return(header("200", $mimehash{$2}, $buf));
    } else {
      FileError("403", "Forbidden", "You don't have permission to access /$File on this server.");
    }
  } else {
    FileError("404", "Not Found", "The requested URL /$File was not found on this server.");	
  }
}

#############################################################################
# autotimer functions
#############################################################################

sub AT_Read {
  my(@at);
  if(-e $AT_FILENAME) {
    open(AT_FILE, $AT_FILENAME) ||
      HTMLError("Cant open $AT_FILENAME!");
    while(<AT_FILE>) {
      chomp;
      next if($_ eq "");
      my($active, $pattern, $section, $start, $stop, $episode, $prio, $lft, $channel) = split(/\:/, $_);
      push(@at, {
				active   => $active,
				pattern  => $pattern,
				section  => $section,
				start    => $start,
				stop     => $stop,
				episode  => $episode,
				prio     => $prio,
				lft      => $lft,
				channel  => $channel
      });
    }
    close(AT_FILE);
  }
  return(@at);
}

sub AT_Write {
  my @at = @_;
  open(AT_FILE, ">" . $AT_FILENAME) || 
    HTMLError("Cant open $AT_FILENAME!");
  for(@at) {
    my $temp;
    for my $item (qw(active pattern section start stop episode prio lft channel)) {
      if(length($temp) == 0) {
				$temp = $_->{$item};
      } else {
				$temp .= ":" . $_->{$item};
      }
    }
    print AT_FILE $temp, "\n";
  }
  close(AT_FILE);
}

sub AutoTimer {
  return if(!$CONFIG{AT_FUNC});
  Log(LOG_AT, "Auto Timer: Scanning for events...");
  my($search, $start, $stop) = @_;
 
  my(@at) = AT_Read();
  
  for my $sender (keys(%EPG)) {
    for my $event (@{$EPG{$sender}}) {
      for my $at (@at) {
				if(!$at->{active}) {
					next;
				}
				
				if($at->{channel}) {
					if($at->{channel} != $event->{number}) {
						next;
					}
				}

				#
				my @section;
				if($at->{section} & 1) {
					push(@section, "title");
				} elsif($at->{section} & 2) {
					push(@section, "subtitle");
				} elsif($at->{section} & 4) {
					push(@section, "description");
				}
				
				#
				next if(!$at->{pattern});
						
				#
				my $fs = 0;
				for my $section (@section) {
					my $fp = 1;
					for my $pattern (split(/ +/, $at->{pattern})) {
						if($event->{$section} !~ /$pattern/i) {
							$fp = 0;
						}
					}
					next if(!$fp);
					$fs = 1;
				}	
				next if(!$fs);
				
				#
				if($at->{start}) {
					if(((my_strftime("%H%M", $event->{start})) < $at->{start})) {
						if(my_strftime("%H%M", $at->{stop}) > my_strftime("%H%M", $at->{start})) {
							next;
						}
					}
				}

				#
				if($at->{stop}) {
					($at->{stop} < $at->{start}) ? $at->{stop} += 2400 : undef;
					if((my_strftime("%H%M", $event->{stop})) > $at->{stop}) {
						next;
					}
				}
				
				#print($event->{title} . " " . (strftime("%H:%M", localtime($event->{time}))) . " " . (strftime("%H:%M", localtime($event->{time} + $event->{duration}))) . "\n");
				
				if($at->{episode}) {
					ProgTimer(1, $event->{vdr_id}, $event->{start}, $event->{stop}, 
						$event->{title} . "~" . $event->{subtitle}, "", $at->{prio}, $at->{lft});
				} else {
					ProgTimer(1, $event->{vdr_id}, $event->{start}, $event->{stop}, 
						$event->{title}, "", $at->{prio}, $at->{lft});
				}
      }
    }
  }
  Log(LOG_AT, "Auto Timer: Done.");
}


#############################################################################
# regulary timers
#############################################################################

sub ProgTimer {
  my($active, $cnumber, $start, $stop, $name, $summary, $prio, $lft) = @_;
  $name =~ s/\:/ /g;
  
  $start -= ($CONFIG{TM_MARGIN_BEGIN} * 60);
  $stop += ($CONFIG{TM_MARGIN_END} * 60);
  
  ($prio = $CONFIG{AT_PRIORITY}) if(!$prio);
  ($lft  = $CONFIG{AT_LIFETIME}) if(!$lft);

  my($newt) = "$active:$cnumber:" . 
    RemoveLeadingZero(strftime("%d", localtime($start))) . ":" . 
    strftime("%H%M", localtime($start)) . ":" .
    strftime("%H%M", localtime($stop)) . ":$prio:" .
    "$lft:$name:$summary";
  my $found = 0;
  my $Update = 0;
  for(ParseTimer(1)) {
    if(($_->{cnumber} == $cnumber) && 
      ($_->{dor} == strftime("%d", localtime($start)))) {
      if(($_->{start} == strftime("%H%M", localtime($start))) 
        && ($_->{stop} == strftime("%H%M", localtime($stop)))) {
				$found = 1;
				if(($_->{name} ne $name) || ($_->{summary} ne $summary)) {
					if($_->{active}) {
						$Update = $_->{id};
					}
				}
      }
    }
  }
  
  if($Update) {
    Log(LOG_AT, "Auto Timer: Updating Timer \"$newt\"");
    my($tmp) = SendCMD(1, "modt $Update $newt");
    if($tmp !~ /^\d+ $newt/) {
      Log(LOG_AT, "Auto Timer: Error updating Timer: \"$tmp\"");
    }
  }
  if(!$found) {
    Log(LOG_AT, "Auto Timer: Programming Timer \"$newt\"");
    my($tmp) = SendCMD(1, "newt $newt");
    if($tmp !~ /^\d+ $newt/) {
      Log(LOG_AT, "Auto Timer: Error programming Timer: \"$tmp\"");
    }
  }
  
}

sub ParseTimer {
  my $pc = shift;
  my $tid = shift;
  my $entry = 1;
  
  my @temp;
  for(SendCMD($pc, "lstt")) {
    last if(/^No timers defined/);
    chomp;
    my($id, $temp) = split(/ /, $_, 2);
    my($active, $vdr_id, $dor, $start, $stop, $prio, $lft, $title, $summary) = split(/\:/, $temp, 9);

    my($startsse, $stopsse, $weekday, $off, $perrec, $length, $first);
    if(length($dor) == 7) { # repeating timer
      $startsse = timelocal(undef, substr($start, 2, 2), substr($start, 0, 2), 
		my_strftime("%d"), (my_strftime("%m") - 1), my_strftime("%Y"));
      $stopsse  = timelocal(undef, substr($stop, 2, 2), substr($stop, 0, 2), 
		my_strftime("%d"), (my_strftime("%m") - 1), my_strftime("%Y"));
      if ($stopsse < $startsse)  {
        $stopsse += 86400;
      }
      $weekday = ((localtime(time))[6] + 6) % 7;
      $perrec = join("", substr($dor, $weekday), substr($dor, 0, $weekday));
      $perrec =~ m/^-+/g;

      $off = (pos $perrec) * 86400;
      if ($off == 0 && $stopsse < time) {
        $weekday = ($weekday + 1) % 7;
        $perrec = join("", substr($dor, ($weekday + 1) % 7), substr($dor, 0, ($weekday + 1) % 7));
        $perrec =~ m/^-+/g;
        $off = ((pos $perrec) + 1) * 86400;
      }
      $startsse += $off;
      $stopsse += $off;
    } elsif(length($dor) == 18) { # first-day timer
      $dor =~ /.{7}\@(\d\d\d\d)-(\d\d)-(\d\d)/;
      $startsse = timelocal(undef, substr($start, 2, 2), 
				substr($start, 0, 2), $3, ($2 - 1), $1);
      $stopsse = timelocal(undef, substr($stop, 2, 2), 
        substr($stop, 0, 2), $stop > $start ? $3 : $3 + 1, 
        ($2 - 1), $1);
    } else { # regular timer
      $startsse = timelocal(undef, substr($start, 2, 2), 
				substr($start, 0, 2), $dor, (my_strftime("%m") - 1),
				my_strftime("%Y"));
			
      $stopsse = timelocal(undef, substr($stop, 2, 2), 
        substr($stop, 0, 2), $stop > $start ? $dor : $dor + 1, 
        (my_strftime("%m") - 1), my_strftime("%Y"));
    }
		
    # move timers which have expired one month into the future
    if(length($dor) != 7 && $stopsse < time) {
      $startsse = timelocal(undef, substr($start, 2, 2), 
        substr($start, 0, 2), $dor, (my_strftime("%m") % 12), 
        my_strftime("%Y") + (my_strftime("%m") == 12 ? 1 : 0));

      $stopsse = timelocal(undef, substr($stop, 2, 2), 
        substr($stop, 0, 2), $stop > $start ? $dor : $dor + 1, 
        (my_strftime("%m") % 12), 
        my_strftime("%Y") + (my_strftime("%m") == 12 ? 1 : 0));
    }

    if($CONFIG{RECORDINGS} && length($dor) == 7)  { # repeating timer
		# generate repeating timer entries for up to 28 days
			$first = 1;
			for($weekday += $off / 86400, $off = 0; $off < 28; $off++)  {
				$perrec = join("", substr($dor, ($weekday + $off) % 7), substr($dor, 0, ($weekday + $off) % 7));
				$perrec =~ m/^-+/g;
				if ((pos $perrec) != 0)  {
					next;
				}

        $length = push(@temp, {
					id        => $id,
					vdr_id    => $vdr_id,
					start     => $startsse,
					stop      => $stopsse,
					startsse  => $startsse + $off * 86400,
					stopsse   => $stopsse + $off * 86400,
					active    => $active,
					cdesc     => get_name_from_vdrid($vdr_id),
					dor       => $dor,
					prio      => $prio,
					lft       => $lft,
					title     => $title,
					summary   => $summary,
					collision => 0,
					critical  => 0,
					first     => $first
        });
        $first = 0;
      }
    } else {
      $length = push(@temp, {
				id        => $id,
				vdr_id    => $vdr_id,
				start     => $startsse,
				stop      => $stopsse,
				startsse  => $startsse,
				stopsse   => $stopsse,
				active    => $active,
				cdesc     => get_name_from_vdrid($vdr_id),
				dor       => $dor,
				prio      => $prio,
				lft       => $lft,
				title     => $title,
				summary   => $summary,
				collision => 0,
				critical  => 0
      });
    }
		
    # save index of entry with specific timer id for later use
    if($tid == $id) {
      $entry = $length;
    }
  }
  
  if($tid) {
    return($temp[$entry - 1]);
  } else {
    return(@temp);
  }
}

#############################################################################
# Tools
#############################################################################

sub getPID {
	open(PID, shift);
	$_ = <PID>;
	close(PID);
	return($_);
}

sub UptoDate {
	my $force = shift;
  if(((time() - $CONFIG{CACHE_LASTUPDATE}) >= ($CONFIG{CACHE_TIMEOUT} * 60)) || $force) {
    ChanTree();
    EPG_buildTree();
    AutoTimer();
    $CONFIG{CACHE_LASTUPDATE} = time();
    CloseSocket();
  }
  return(0);
}

sub Log {
  if($#_ >= 1) {
		if($CONFIG{LOGGING}) {
			my $level = shift;
			my $message = join("", @_);
			if($CONFIG{LOGLEVEL} & $level) {
				open(LOGFILE, ">>" . $LOGFILE);
				print LOGFILE "$message\n";
				close(LOGFILE);
			}
		}
  } else {
    print STDERR "bogus Log() call\n";
  }
}

sub TemplateNew {
  my $file = shift;
  $CONFIG{LANGUAGE} = "English" if(!$CONFIG{LANGUAGE});
  my $template = HTML::Template->new(
    die_on_bad_params => 0,
    loop_context_vars => 1,
    filename => "$TEMPLATEDIR/$CONFIG{LANGUAGE}/$file");
  return $template;
}

sub my_strftime {
  my $format = shift;
  my $time = shift;
  return(strftime($format, $time ? localtime($time) : localtime(time)));
}

sub GetFirstChannel {
	return($CHAN[0]->{service_id});
}

sub ChannelHasEPG {
	my $service_id = shift;
	for my $event (@{$EPG{$service_id}}) {
		return(1);
	} 
	return(0);
}

sub Encode_Referer {
  return(MIME::Base64::encode_base64(shift));
}

sub Decode_Referer {
  return(MIME::Base64::decode_base64(shift));
}

sub encode_ref {
  my($tmp) = $_[0]->url(-relative=>1,-query=>1);
  my(undef, $query) = split(/\?/, $tmp, 2);
  return(MIME::Base64::encode_base64($query));
}

sub decode_ref {
  return(MIME::Base64::decode_base64($_[0]));
}

sub access_log {
	my $ip               = shift;
	my $username         = shift;
	my $time             = shift;
	my $rawrequest       = shift;
	my $http_status      = shift;
	my $bytes_transfered = shift;
	my $request          = shift;
	my $useragent        = shift;
	return sprintf("%s - %s [%s +0100] \"%s\" %s %s \"%s\" \"%s\"",
		$ip,
		$username,
		my_strftime("%d/%b/%Y:%H:%M:%S", $time),
		$rawrequest,
		$http_status,
		$bytes_transfered,
		$request,
		$useragent
	);
}

sub ReadConfig {    
  if(-e $CONFFILE) {
    open(CONF, $CONFFILE);
    while(<CONF>) {
      chomp;
      my($key, $value) = split(/ \= /, $_, 2);
      $CONFIG{$key} = $value;
    }
    close(CONF);
  } else {
    open(CONF, ">$CONFFILE");
    for(keys(%CONFIG)) {
      print(CONF "$_ = $CONFIG{$_}\n");
    }
    close(CONF);
    return(1);
  }
  return(0);
}   

sub Question {
  my($quest, $default) = @_;
  print("$quest [$default]: ");
  my($answer);
  $answer = <STDIN>;
  if($answer eq "\n") {
    return($default);
  } else {
    return($answer);
  }
}

sub RemoveLeadingZero {
  my($str) = @_;
  while(substr($str, 0, 1) == 0) {
    $str = substr($str, 1, (length($str) - 1));
  }
  return($str);
}

sub csvAdd {
  my $csv = shift;
  my $add = shift;
  
  my $found = 0;
  for my $item (split(",", $csv)) {
		$found = 1 if($item eq $add);
  }
	$csv = join(",", (split(",", $csv), $add)) if(!$found);
  return($csv);
}
  
sub csvRemove {
  my $csv = shift;
  my $remove = shift;
  
  my $newcsv;
  for my $item (split(",", $csv)) {
    if($item ne $remove) {
      my $found = 0;
      for my $dup (split(",", $newcsv)) {
				$found = 1 if($dup eq $item);
      }
			$newcsv = join(",", (split(",", $newcsv), $item)) if(!$found);
    }
  }
  return($newcsv);
}

#############################################################################
# frontend
#############################################################################
##
#
##
sub prog_detail {
  return if(UptoDate());
	my $vdr_id = $q->param("vdr_id");
	my $epg_id = $q->param("epg_id");
	
  my($titel, $text);
	
	if($vdr_id && $epg_id) {
		for(@{$EPG{$vdr_id}}) {
			if($_->{id} == $epg_id) {
				$titel = $_->{progname} . ' | ' . my_strftime("%H:%M", $_->{start}) . '-' . my_strftime("%H:%M", $_->{stop}) . ' | ' . $_->{title};
				$text = $_->{summary};
				last;
			}
		}
	}
	my $template = TemplateNew("prog_detail.html");
  $template->param(
    titel => $titel ? $titel : undef, 
    text  => $text ? $text : undef,
  );
  return(header("200", "text/html", $template->output)); 
}


##
#
##
sub prog_list {
  return if(UptoDate());
	my $vdr_id = $q->param("vdr_id");
  
	# called without vdr_id, redirect to the first known channel
  if(!$vdr_id) {
    return(headerForward("$MyURL?aktion=prog_list&vdr_id=1"));
  }
	
	#
  my @channel;
	for my $channel (@CHAN) { 
    # if its wished, display only wanted channels
    if($CONFIG{CHANNELS_WANTED_PRG}) {
      my $found = 0;
      for my $n (split(",", $CONFIG{CHANNELS_WANTED})) {
				($found = 1) if($n eq $channel->{vdr_id});
      }
      next if(!$found);
    } 
    
    # skip channels without EPG data
		if(ChannelHasEPG($channel->{vdr_id})) {
			push(@channel, {
				name       => $channel->{name},
				vdr_id     => $channel->{vdr_id},
				current    => ($vdr_id == $channel->{vdr_id}) ? 1 : 0,
      }); 
    }
  }
  
  # find the next/prev channel
  my $ci = 0;
  for(my $i = 0; $i <= $#channel; $i++) {
    ($ci = $i) if($vdr_id == $channel[$i]->{vdr_id});
  }
  my ($next_channel, $prev_channel);
  ($prev_channel = $channel[$ci - 1]->{vdr_id}) if($ci > 0);
  ($next_channel = $channel[$ci + 1]->{vdr_id}) if($ci < $#channel); 
	
	#
  my(@show, $progname, $cnumber); my($day) = 0;
  for my $event (@{$EPG{$vdr_id}}) {
    if(my_strftime("%d", $event->{start}) != $day) {
      push(@show, {
				title  => $event->{channel_name} . " | " . 
					FullDay(my_strftime("%w", $event->{start})) . ", " . 
					my_strftime("%d.%m.%Y", $event->{start}),
				newd   => 1,
				next_channel => $next_channel ? "$MyURL?aktion=prog_list&vdr_id=$next_channel" : undef,
				prev_channel => $prev_channel ? "$MyURL?aktion=prog_list&vdr_id=$prev_channel" : undef,
      });
      $day = strftime("%d", localtime($event->{start}));
    }
    push(@show, {
      ssse        => $event->{start},
      emit        => my_strftime("%H:%M", $event->{start}),
      duration    => my_strftime("%H:%M", $event->{stop}),
      title       => $event->{title},
      subtitle    => $event->{subtitle},
      recurl      => "$MyURL?aktion=timer_new_form&epg_id=" . $event->{id} . "&vdr_id=" . $event->{vdr_id},
      infurl      => $event->{summary} ? $MyURL . "?aktion=prog_detail&epg_id=" . $event->{id} . "&vdr_id=" . $event->{vdr_id} : undef,
      newd        => 0,
    });
    $progname = $event->{progname};
    $cnumber = $event->{cnumber};
  }
	
  # 
  my($template) = TemplateNew("prog_list.html");
  $template->param(
    url      => $MyURL,
    loop     => \@show,
    chanloop => \@channel,
    progname => GetChannelDescByNumber($vdr_id),
		switchurl=> "$MyURL?aktion=prog_switch&channel=" . $vdr_id,
  );
  return(header("200", "text/html", $template->output));
}


##
#
##
sub timer_list {
  return if(UptoDate()); 
	
	ParseTimer(0);
  my @timer;
  for my $timer (ParseTimer(0)) {
    $timer->{delurl} = $MyURL . "?aktion=timer_delete&timer_id=" . $timer->{id},
    $timer->{modurl} = $MyURL . "?aktion=timer_new_form&timer_id=" . $timer->{id},
    $timer->{start} = my_strftime("%H:%M", $timer->{start});
    $timer->{stop} = my_strftime("%H:%M", $timer->{stop});
    push(@timer, $timer);
  }
  @timer = sort({ $a->{startsse} <=> $b->{startsse} } @timer);
	
  if ($CONFIG{RECORDINGS})  {
    my($ii, $jj, $timer, $last);
    for ($ii = 0; $ii < @timer; $ii++)  {
      if (length($timer[$ii]->{dor}) != 7 || $timer[$ii]->{first})  {
        $last = $ii;
      }
      for ($jj = 0; $jj < $ii; $jj++)  {
        if ($timer[$ii]->{startsse} >= $timer[$jj]->{startsse} &&
            $timer[$ii]->{startsse} < $timer[$jj]->{stopsse}) {
          if ($timer[$ii]->{active} && $timer[$jj]->{active}) { 
            ($timer[$ii]->{collision})++; 
          }
        }
      }
    }

    splice(@timer, $last + 1);

    for ($ii = 0; $ii < @timer; $ii++)  {
      $timer[$ii]->{critical} = $timer[$ii]->{collision} >= $CONFIG{RECORDINGS}; 
      $timer[$ii]->{collision} = $timer[$ii]->{collision} >= ($CONFIG{RECORDINGS} - 1); 
    }
  }
  
  #
  my(@at); 
  my($id) = 0;
  for(AT_Read()) {
    $id++;
    if($_->{start}) {
      $_->{start} = substr($_->{start}, 0, 2) . ":" . substr($_->{start}, 2, 5);
    }
    if($_->{stop}) {
      $_->{stop} = substr($_->{stop}, 0, 2) . ":" . substr($_->{stop}, 2, 5);
    }
    $_->{modurl} = $MyURL . "?aktion=at_timer_edit&id=$id";
    $_->{delurl} = $MyURL . "?aktion=at_timer_delete&id=$id";
    $_->{prio} = $_->{prio} ? $_->{prio} : $CONFIG{AT_PRIORITY};
    $_->{lft} = $_->{lft} ? $_->{lft} : $CONFIG{AT_LIFETIME};
    $_->{id} = $id;
    $_->{channel} = GetChannelDescByNumber($_->{channel}); 
    push(@at, $_);
  }
  @at = sort({ lc($a->{pattern}) cmp lc($b->{pattern}) } @at);
  
	my $template = TemplateNew("timer_list.html");
  $template->param(
    timer_loop => \@timer,
    at_timer_loop => \@at,
    at_timer_func => $CONFIG{AT_FUNC},
    nturl  => $MyURL . "?aktion=timer_new_form",
    naturl => $MyURL . "?aktion=at_timer_new&active=1",
    url    => $MyURL,
  );
  return(header("200", "text/html", $template->output));
}


##
#
##
sub at_timer_edit {
  my $id = $q->param("id");

  my @at = AT_Read();
  
  #
  my @chans;
  for my $chan (@CHAN) {
    if($chan->{number}) {
      $chan->{cur} = ($chan->{number} == $at[$id-1]->{channel}) ? 1 : 0;
      push(@chans, $chan);
    }
  }
  
  my $template = TemplateNew("at_new.html");
  $template->param(
    channels=> \@chans, 
    id      => $id,
    url     => $MyURL,
    prio    => $at[$id-1]->{prio} ? $at[$id-1]->{prio} : $CONFIG{AT_PRIORITY},
    lft     => $at[$id-1]->{lft} ? $at[$id-1]->{lft} : $CONFIG{AT_LIFETIME},
    active  => $at[$id-1]->{active},
    episode => $at[$id-1]->{episode},
    pattern => $at[$id-1]->{pattern},
    starth  => (length($at[$id-1]->{start}) >= 4) ? 
      substr($at[$id-1]->{start}, 0, 2) : undef,
    startm  => (length($at[$id-1]->{start}) >= 4) ?
      substr($at[$id-1]->{start}, 2, 5) : undef,
    stoph   => (length($at[$id-1]->{stop}) >= 4) ?
      substr($at[$id-1]->{stop}, 0, 2) : undef,
    stopm   => (length($at[$id-1]->{stop}) >= 4) ?
      substr($at[$id-1]->{stop}, 2, 5) : undef,
    title   => ($at[$id-1]->{section} & 1) ? 1 : 0,
    subtitle=> ($at[$id-1]->{section} & 2) ? 1 : 0,
    description => ($at[$id-1]->{section} & 4) ? 1 : 0,
  );
  return(header("200", "text/html", $template->output));
}


##
#
##
sub at_timer_new {
	my $template = TemplateNew("at_new.html");
  $template->param(
    url      => $MyURL, 
    active   => $q->param("active"),
    title    => 1,
    channels => \@CHAN,
    prio     => $CONFIG{AT_PRIORITY},
    lft      => $CONFIG{AT_LIFETIME},
  );
  return(header("200", "text/html", $template->output));
};


##
#
##
sub at_timer_save {
  my $id = $q->param("id");
  
  if($q->param("save")) {
    if(!$id) {
      my @at = AT_Read();
      my $section = 0;
      ($section += 1) if($q->param("title"));
      ($section += 2) if($q->param("subtitle"));
      ($section += 4) if($q->param("description"));
      push(@at, {
	episode  => $q->param("episode") ? $q->param("episode") : 0,
	active   => $q->param("active"),
	pattern  => $q->param("pattern"),
	section  => $section,
	start    => $q->param("starth") . $q->param("startm"),
	stop     => $q->param("stoph") . $q->param("stopm"),
	prio     => $q->param("prio"),
	lft      => $q->param("lft"),
	channel  => $q->param("channel"),
      });
      AT_Write(@at);
    } else {
      my $z = 0;
      my @at;
      for(AT_Read()) {
	$z++;
	if($z != $id) {
	  push(@at, $_);
	} else { 
	  my $section = 0;
	  ($section += 1) if($q->param("title"));
	  ($section += 2) if($q->param("subtitle"));
	  ($section += 4) if($q->param("description"));
	  push(@at, {
	    episode  => $q->param("episode") ? $q->param("episode") : 0,
	    active   => $q->param("active"),
	    pattern  => $q->param("pattern"),
	    section  => $section,
	    start    => $q->param("starth") . $q->param("startm"),
	    stop     => $q->param("stoph") . $q->param("stopm"),
	    prio     => $q->param("prio"),
	    lft      => $q->param("lft"),
	    channel  => $q->param("channel"),
	  });
	}
      }
      AT_Write(@at);
    }
    
    $CONFIG{LOCALLASTGET} = 0;
    UptoDate();
  }
  headerForward("$MyURL?aktion=timer_list");
}


##
#
##
sub at_timer_delete {
  my $id = $q->param("id");
  my $z = 0;

  my @at = AT_Read();
  my @new;
  if($id) { 
    for(@at) {
      $z++;
      push(@new, $_) if($id != $z);
    }
    AT_Write(@new);
  } else {
    my @sorted;
    for($q->param) {
      if(/xxxx_(.*)/) {
				push(@sorted, $1);
      } 
    }
   @sorted = sort({ $b <=> $a } @sorted);
    my $z = 0;
    for my $at (@at) {
      $z++;
      my $push = 1;
      for my $sorted (@sorted) {
				($push = 0) if($z == $sorted);
      }
      push(@new, $at) if($push);
    }
    AT_Write(@new); 
  }
  headerForward("$MyURL?aktion=timer_list");
}


##
#
##
sub timer_new_form {
  UptoDate();
  
	my $epg_id   = $q->param("epg_id");
	my $vdr_id   = $q->param("vdr_id");
  my $timer_id = $q->param("timer_id");
  
  my $this_event;
  if($epg_id) { # new timer
    $this_event = EPG_getEntry($vdr_id, $epg_id);
    $this_event->{active} = 1;
    $this_event->{start} -= ($CONFIG{TM_MARGIN_BEGIN} * 60);
    $this_event->{stop}  += ($CONFIG{TM_MARGIN_END}  * 60);
  } elsif($timer_id) { # edit existing timer
    $this_event = ParseTimer(0, $timer_id);
  } else { # none of the above
    $this_event->{start}   = time();
    $this_event->{stop}    = 0;
    $this_event->{active}  = 1;
    $this_event->{vdr_id}  = 1;
  }
  
  my @channels;
	for my $channel (@CHAN) {
    ($channel->{vdr_id} == $this_event->{vdr_id}) ? ($channel->{current} = 1) : ($channel->{current} = 0);
    push(@channels, $channel);
  }
  
  my $template = TemplateNew("timer_new.html");
  $template->param(
    url      => $MyURL,
    active   => $this_event->{active},
    starth   => my_strftime("%H", $this_event->{start}),
    startm   => my_strftime("%M", $this_event->{start}),
    stoph    => $this_event->{stop} ? my_strftime("%H", $this_event->{stop}) : "00",
    stopm    => $this_event->{stop} ? my_strftime("%M", $this_event->{stop}) : "00",
    dor      => (length($this_event->{dor}) == 7) ? $this_event->{dor} : my_strftime("%d", $this_event->{start}),
    prio     => $this_event->{prio} ? $this_event->{prio} : $CONFIG{TM_PRIORITY},
    lft      => $this_event->{lft}  ? $this_event->{lft}  : $CONFIG{TM_LIFETIME},
    title    => $this_event->{title},
    summary  => $this_event->{summary},
    timer_id => $timer_id ? $timer_id : 0,
    referer  => 0,
    channels => \@channels,
    newtimer => $timer_id ? 0 : 1,
    referer  => Encode_Referer($Referer),
  );
  return(header("200", "text/html", $template->output));
}


##
#
##
sub timer_add {
  #XXX
  my $timer_id  = $q->param("timer_id");
  
	my $data;
  
	if($q->param("save")) {
  
		if($q->param("starth") =~ /\d+/ && $q->param("starth") < 24 && $q->param("starth") >= 0) {
			$data->{start} = $q->param("starth");
		} else { print "Help!\n"; }
		if($q->param("startm") =~ /\d+/ && $q->param("startm") < 60 && $q->param("startm") >= 0) {
			$data->{start} .= $q->param("startm");
		} else { print "Help!\n"; }

		if($q->param("stoph") =~ /\d+/ && $q->param("stoph") < 24 && $q->param("stoph") >= 0) {
			$data->{stop} = $q->param("stoph");
		} else { print "Help!\n"; }
		if($q->param("stopm") =~ /\d+/ && $q->param("stopm") < 60 && $q->param("stopm") >= 0) {
			$data->{stop} .= $q->param("stopm");
		} else { print "Help!\n"; }
	 
		if($q->param("prio") =~ /\d\d/) {
			$data->{prio} = $q->param("prio");
		}

		if($q->param("lft") =~ /\d\d/) {
			$data->{lft} = $q->param("lft");
		}
		
		if($q->param("active") == 0 || $q->param("active") == 1) {
			$data->{active} = $q->param("active");
		}
		
		if($q->param("dor") =~ /[0-9MTWTWFSS-]+/) {
			$data->{dor} = $q->param("dor");
		}
		
		if($q->param("channel") =~ /\d+/) {
			$data->{channel} = $q->param("channel");
		}

		if(length($q->param("title")) > 0) {
			$data->{title} = $q->param("title");
			$data->{title} =~ s/\://g;
		}

		if(length($q->param("summary")) > 0) {
			$data->{summary} = $q->param("summary");
			$data->{summary} =~ s/\://g;
			$data->{summary} =~ s/[\r\n]//g;
		}
		
		my $return = SendCMD(0, 
			sprintf("%s %s:%s:%s:%s:%s:%s:%s:%s:%s",
				$timer_id ? "modt $timer_id" : "newt",
				$data->{active},
				$data->{channel},
				$data->{dor},
				$data->{start},
				$data->{stop},
				$data->{prio},
				$data->{lft},
				$data->{title},
				$data->{summary}
			)
		);
	}

  if($q->param("referer")) {
    return(headerForward(Decode_Referer($q->param("referer"))));
  } else {
    return(headerForward("$MyURL?aktion=timer_list"));
  }
}


##
#
##
sub timer_delete {
  my($timer_id) = $q->param('timer_id');
  if($timer_id) {
    my($result) = SendCMD(0, "delt $timer_id");
    if($result =~ /Timer "$timer_id" is recording/i) {
      SendCMD(0, "modt $timer_id off");
    }
  } else {
    my @sorted;
    for($q->param) {
      if(/xxxx_(.*)/) {
				push(@sorted, $1);
      }
    }
    @sorted = sort({ $b <=> $a } @sorted);
    for my $t (@sorted) {
      my($result) = SendCMD(1, "delt $t");
      if($result =~ /Timer "$t" is recording/i) {
				SendCMD(1, "modt $t off");
      }
    }
    CloseSocket();
  }
  headerForward("$MyURL?aktion=timer_list");
}


##
#
##
sub prog_summary {
  return if(UptoDate());
  my $time = $q->param("time");
  my $search = $q->param("search");
  
  # zeitpunkt bestimmen
  my $event_time;
  if($time ne "") {
    my ($hour, $minute);
    if($time =~ /(\d{1,2})(\D?)(\d{1,2})/) {
      if(length($1) == 1 && length($3) == 1 && !$2) {
				$hour = $1 . $3;
      } else {
				($hour, $minute) = ($1, $3);
      }
    } elsif($time =~ /\d/) {
      $hour = $time;
    }

    if($hour <= my_strftime("%H") && $minute < my_strftime("%M")) {
      $event_time = timelocal(
				0, 
				$minute, 
				$hour, 
				my_strftime("%d", time + 86400), 
				(my_strftime("%m", time + 86400) - 1),
				my_strftime("%Y")
			) + 1;
    } else {
      $event_time = timelocal(
				0, 
				$minute, 
				$hour, 
				my_strftime("%d"), 
				(my_strftime("%m") - 1),
				my_strftime("%Y")
			) + 1;
    } 
  } else {
    $event_time = time();
  }
	
  my(@show, @shows, @temp);
  for(keys(%EPG)) {
    for my $event (@{$EPG{$_}}) {
      if(!$search) {
				if($CONFIG{CHANNELS_WANTED_SUMMARY}) {
					my $f = 0;
					for my $n (split(/\,/, $CONFIG{CHANNELS_WANTED})) {
						($f = 1) if($n eq $event->{vdr_id});
					}
				 next if(!$f);
				}
				next if($event_time > $event->{stop});
			} else {
				my($found);
				for my $word (split(/ +/, $search)) {
					$found = 0;
					for my $section (qw(title subtitle description)) {
						if($event->{$section} =~ /$word/i) {
							$found = 1;
						}
					}
					if(!$found) {
						last;
					}
				}
				next if(!$found);
			}

      push(@show,  {
				date     => my_strftime("%d.%m.%Y", $event->{start}),
				start    => my_strftime("%H:%M", $event->{start}),
				stop     => my_strftime("%H:%M", $event->{stop}),
				title    => $event->{title},
				subtitle => length($event->{subtitle}) > 30 ? substr($event->{subtitle}, 0, 30) . "..." : $event->{subtitle},
				progname => $event->{channel_name},
				summary  => length($event->{summary}) > 120 ? substr($event->{summary}, 0, 120) . "..." : $event->{summary},
				vdr_id   => $event->{vdr_id},
				proglink => "$MyURL?aktion=prog_list&number=" . $event->{vdr_id},
				switchurl=> $MyURL . "?aktion=prog_switch&channel=" . $event->{vdr_id},
				infurl   => $event->{summary} ? "$MyURL?aktion=prog_detail&epg_id=" . $event->{id} . "&vdr_id=" . $event->{vdr_id} : undef,
				recurl   => "$MyURL?aktion=timer_new_form&epg_id=" . $event->{id} . "&vdr_id=" . $event->{vdr_id},
      });
      last if(!$search);
    }
  }
	
	# needed for vdr 1.0.x, dunno why
  @show = sort({ $a->{vdr_id} <=> $b->{vdr_id} } @show);

	#
	my $spalten = 4;
  for(my $i = 0; $i <= $#show; $i++) {
    undef(@temp);
    for(my $z = 0; $z < $spalten; $i++, $z++) {
      push(@temp, $show[$i]);
    }
    $i--;
    push(@shows, { day => [ @temp ] });
  }
	
  #
	my $template = TemplateNew("prog_summary.html");
  $template->param(
    rows   => \@shows,
    now    => strftime("%H:%M", localtime($event_time)),
    nowurl => $MyURL . "?aktion=prog_summary",
    url    => $MyURL
  );
  return(header("200", "text/html", $template->output));
}


#
sub rec_list {	
  my @recordings;
  my($sortby) = $q->param('sortby');
  my(@response) = SendCMD(0, "lstr");
  for(@response) {
    chomp;
    next if(length($_) == 0);
    if(/^No recordings available/) {
      last;
    }
    my($new);
    my($id, $date, $time, $name) = split(/ +/, $_, 4);
    
    #
    if(substr($time, 5, 1) eq "*") {
      $new = 1;
      $time =~ s/\*//;
    }
		
    push(@recordings, {
      sse    => timelocal(undef, substr($time, 3, 2), 
					substr($time, 0, 2), substr($date, 0, 2),
					(substr($date, 3, 2) - 1),
					my_strftime("%Y")),
      date   => $date,
      time   => $time,
      name   => $name,
      new    => $new,
      id     => $id,
      delurl => $MyURL . "?aktion=rec_delete&id=$id",
      infurl => $MyURL . "?aktion=rec_detail&id=$id"
    });
  }
	
	my $desc;
	$q->param("desc") ? ($desc = 1) : ($desc = 0);
	
  if($sortby eq "time") {
		if(!$desc) {
			@recordings = sort({ $a->{time} cmp $b->{time} } @recordings);
		} else {
			@recordings = sort({ $b->{time} cmp $a->{time} } @recordings);
		}
  } elsif($sortby eq "name") {
		if(!$desc) {
			@recordings = sort({ lc($a->{name}) cmp lc($b->{name}) } @recordings);
		} else {
			@recordings = sort({ lc($b->{name}) cmp lc($a->{name}) } @recordings);
		}
  } else {
		if(!$desc) {
			@recordings = sort({ $a->{sse} cmp $b->{sse} } @recordings);
		} else {
			@recordings = sort({ $b->{sse} cmp $a->{sse} } @recordings);
		}
  }
	
	$desc ? ($desc = 0) : ($desc = 1);
	
	my $template = TemplateNew("rec_list.html");
  $template->param(
    recloop    => \@recordings,
    sortbydate => "$MyURL?aktion=rec_list&desc=$desc",
    sortbytime => "$MyURL?aktion=rec_list&sortby=time&desc=$desc",
    sortbyname => "$MyURL?aktion=rec_list&sortby=name&desc=$desc",
  );
  return(header("200", "text/html", $template->output));
}


##
#
##
sub rec_delete {
  my($id) = $q->param('id');
  if($id) {
    SendCMD(1, "delr $id");
  } else {
    for($q->param) {
      if(/xxxx_(.*)/) {
				SendCMD(1, "delr $1");
      }
    }
  }
  CloseSocket();
  headerForward("$MyURL?aktion=rec_list");
}


##
#
##
sub show_top {
	my $template = TemplateNew("top.html");
  return(header("200", "text/html", $template->output));
}


##
#
##
sub rec_detail {
  my($id) = $q->param('id');

  #
  my($i, $titel);
  for(SendCMD(0, "lstr")) {
    ($i, undef, undef, $titel) = split(/ +/, $_, 4);
    last if($id == $i);
  }
  chomp($titel);

  #
  my($text); my($first) = 1;
  my($result) = SendCMD(0, "lstr $id");
  if($result !~ /No summary availab/i) {
    for(split(/\|/, $result)) {
      if($_ ne (split(/\~/, $titel))[1]) {
	if($first && $titel !~ /\~/ && length($titel) < 20) {
	  $titel .= "~" . $_;
	  $first = 0;
	} else {
	  $text .= $_;
	}
      } 
    }
  }

  #
  $titel =~ s/\~/ - /;
  
	my $template = TemplateNew("prog_detail.html");
  $template->param(
    text  => $text ? $text : "",
    titel => $titel
  );
  return(header("200", "text/html", $template->output));	
}


##
#
##
sub conf_list {  
  return if(UptoDate());
  
	sub ApplyConfig {
    for($q->param) {
      if(/[A-Z]+/) {
				$CONFIG{$_} = $q->param($_);
      }
    }
  }

	sub WriteConfig {
		open(CONF, ">$CONFFILE");
		for my $key (sort(keys(%CONFIG))) {
			print CONF "$key = $CONFIG{$key}\n";
		}
		close(CONF);
	}
  
  if($q->param("submit") eq ">>>>>") {
    for my $vdr_id ($q->param("all_channels")) {
      $CONFIG{CHANNELS_WANTED} = csvAdd($CONFIG{CHANNELS_WANTED}, $vdr_id); 
    }
    ApplyConfig(); WriteConfig();
  } elsif($q->param("submit") eq "<<<<<") {
    for my $vdr_id ($q->param("selected_channels")) {
      $CONFIG{CHANNELS_WANTED} = csvRemove($CONFIG{CHANNELS_WANTED}, $vdr_id); 
    }
    ApplyConfig(); WriteConfig();
  } elsif($q->param("save")) {
    ApplyConfig(); WriteConfig(); 
  } elsif($q->param("apply")) {
    ApplyConfig();
  }
  
  #
  my @lang;
  for my $dir (<$TEMPLATEDIR/*>) {
    next if(!-d $dir);
    $dir =~ s/.*\///g;
    my $found = 0;
    for(@lang) { ($found = 1) if($_->{name} eq $1); }
    if(!$found) { 
      push(@lang, {
				name   => $dir,
				aklang => ($CONFIG{LANGUAGE} eq $dir) ? 1 : 0,
      }); 
    }
  }
  
  #
  my (@all_channels, @selected_channels);
  for my $channel (@CHAN) {
    #
    push(@all_channels, {
      name   => $channel->{name},
      vdr_id => $channel->{vdr_id}
    });
    
    # 
    my $found = 0;
    for(split(",", $CONFIG{CHANNELS_WANTED})) {
      if($_ eq $channel->{vdr_id}) {
				$found = 1;
      }
    }
    next if !$found; 
    push(@selected_channels, {
      name   => $channel->{name},
      vdr_id => $channel->{vdr_id}
    }); 
  }

  my $template = TemplateNew("config.html");
  $template->param(
    %CONFIG,
    LANGLIST          => \@lang,
    ALL_CHANNELS      => \@all_channels,
    SELECTED_CHANNELS => \@selected_channels,
    url               => $MyURL,
  );
  return(header("200", "text/html", $template->output));


}

sub prog_switch {
  my $channel = $q->param("channel");
  if($channel) {
    SendCMD(0, "chan $channel");
  }
  SendFile($BASENAME . "/images/switch_channel.gif");    
}

sub rc_show {
	my $template = TemplateNew("rc.html");
  $template->param(
    surl_0         => $MyURL . "?aktion=rc_hitk&key=0",
    surl_1         => $MyURL . "?aktion=rc_hitk&key=1",
    surl_2         => $MyURL . "?aktion=rc_hitk&key=2",
    surl_3         => $MyURL . "?aktion=rc_hitk&key=3",
    surl_4         => $MyURL . "?aktion=rc_hitk&key=4",
    surl_5         => $MyURL . "?aktion=rc_hitk&key=5",
    surl_6         => $MyURL . "?aktion=rc_hitk&key=6",
    surl_7         => $MyURL . "?aktion=rc_hitk&key=7",
    surl_8         => $MyURL . "?aktion=rc_hitk&key=8",
    surl_9         => $MyURL . "?aktion=rc_hitk&key=9",
    
    surl_power     => $MyURL . "?aktion=rc_hitk&key=Power",
    
    surl_ok        => $MyURL . "?aktion=rc_hitk&key=Ok",

    surl_menu      => $MyURL . "?aktion=rc_hitk&key=Menu",
    surl_back      => $MyURL . "?aktion=rc_hitk&key=Back",
    
    surl_up        => $MyURL . "?aktion=rc_hitk&key=Up",
    surl_down      => $MyURL . "?aktion=rc_hitk&key=Down",
    surl_left      => $MyURL . "?aktion=rc_hitk&key=Left",
    surl_right     => $MyURL . "?aktion=rc_hitk&key=Right",

    surl_red       => $MyURL . "?aktion=rc_hitk&key=Red",
    surl_green     => $MyURL . "?aktion=rc_hitk&key=Green",
    surl_blue      => $MyURL . "?aktion=rc_hitk&key=Blue",
    surl_yellow    => $MyURL . "?aktion=rc_hitk&key=Yellow",
    
    surl_volplus   => $MyURL . "?aktion=rc_hitk&key=Volume%2B",
    surl_volminus  => $MyURL . "?aktion=rc_hitk&key=Volume-",

  );
  return(header("200", "text/html", $template->output));
}

sub rc_hitk {
  my $key = $q->param("key");
  SendCMD(0, "hitk $key");
  SendFile("images/transe.gif");
  #SendHeader("200", "text/html", "0");
}

# EOF
